/* 
 * Copyright 1994 Alastair Gourlay
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice appear in all copies and that both that 
 * copyright notice and this permission notice appear in supporting 
 * documentation.
 */

/*
 *
 * Grid.c - XmpGrid widget
 *
 * XmpGeometry subclass. Simple Grid widget.
 *
 */


#include <stdio.h>
#include "GridP.h"

static void Initialize(Widget, Widget, ArgList, Cardinal*);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal*);
static void ConstraintInitialize(Widget, Widget, ArgList, Cardinal*);
static Boolean ConstraintSetValues(Widget,Widget,Widget,ArgList,Cardinal*);
static void Size(Widget,Dimension*,Dimension*);
static void Resize(Widget);

#define DefaultRows 4
#define DefaultColumns 4

#define RowsZeroMessage "XmNrows must be greater than 0!"
#define ColumnsZeroMessage "XmNcolumns must be greater than 0!"

static XtResource resources[] = 
{
    {
	XmNmarginWidth,
	XmCMarginWidth,
	XmRHorizontalDimension,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpGridRec, grid.margin_width),
	XtRImmediate,
	(XtPointer)2
    },
    {
	XmNmarginHeight,
	XmCMarginHeight,
	XmRVerticalDimension,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpGridRec, grid.margin_height),
	XtRImmediate,
	(XtPointer)2
    },
    {
	XmNrows, 
	XmCRows, 
	XmRShort, 
	sizeof (short),
	XtOffsetOf(struct _XmpGridRec, grid.rows),
	XmRImmediate, (XtPointer)DefaultRows
    },
    {
	XmNcolumns, 
	XmCColumns, 
	XmRShort, 
	sizeof (short),
	XtOffsetOf(struct _XmpGridRec, grid.columns),
	XmRImmediate, (XtPointer)DefaultColumns
    },
};


static XmSyntheticResource syn_resources[] =
{
    { 
	XmNmarginWidth,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpGridRec, grid.margin_width),
	_XmFromHorizontalPixels,
	_XmToHorizontalPixels 
    },
    { 
	XmNmarginHeight,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpGridRec, grid.margin_height),
	_XmFromVerticalPixels,
	_XmToVerticalPixels 
    },
};

static XtResource constraints[] =
{
    {
	XmNgridMarginWidth,
	XmCGridMarginWidth,
	XmRHorizontalDimension,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpGridConstraintRec, grid.grid_margin_width),
	XmRImmediate, (XtPointer) 0
    },
    {
	XmNgridMarginHeight,
	XmCGridMarginHeight,
	XmRVerticalDimension,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpGridConstraintRec, grid.grid_margin_height),
	XmRImmediate, (XtPointer) 0
    },
};

static XmSyntheticResource syn_constraints[] =
{
    {
	XmNgridMarginWidth,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpGridConstraintRec, grid.grid_margin_width),
	_XmFromHorizontalPixels,
	_XmToHorizontalPixels
    },
    {
	XmNgridMarginHeight,
	sizeof (Dimension),
	XtOffsetOf(struct _XmpGridConstraintRec, grid.grid_margin_height),
	_XmFromVerticalPixels,
	_XmToVerticalPixels
    },
};



/****************************************************************
 *
 * XmpGrid class record
 *
 ****************************************************************/

externaldef(xmgridclassrec) XmpGridClassRec xmpGridClassRec = 
{
  {                                           /* core_class            */
    (WidgetClass)&xmpGeometryClassRec,        /* superclass            */
    "XmpGrid",                                /* class_name            */
    sizeof(XmpGridRec),                       /* widget_size           */
    NULL,                                     /* class_initialize      */
    NULL,                                     /* class_part_initialize */
    FALSE,                                    /* class_inited          */
    Initialize,                               /* initialize            */
    NULL,                                     /* initialize_hook       */
    XtInheritRealize,                         /* realize               */
    NULL,                                     /* actions               */
    0,                                        /* num_actions           */
    resources,                                /* resources             */
    XtNumber(resources),                      /* num_resources         */
    NULLQUARK,                                /* xrm_class             */
    TRUE,                                     /* compress_motion       */
    XtExposeCompressMaximal,                  /* compress_exposure     */
    TRUE,                                     /* compress_enterleave   */
    FALSE,                                    /* visible_interest      */
    NULL,                                     /* destroy               */
    Resize,                                   /* resize                */
    XtInheritExpose,                          /* expose                */
    SetValues,                                /* set_values            */
    NULL,                                     /* set_values_hook       */
    XtInheritSetValuesAlmost,                 /* set_values_almost     */
    NULL,                                     /* get_values_hook       */
    NULL,                                     /* accept_focus          */
    XtVersion,                                /* version               */
    NULL,                                     /* callback_private      */
    XtInheritTranslations,                    /* tm_table              */
    XtInheritQueryGeometry,                   /* query_geometry        */
    NULL,                                     /* display_accelerator   */
    NULL,                                     /* extension             */
  },

  {                                           /* composite_class       */
    XtInheritGeometryManager,                 /* geometry_manager      */
    XtInheritChangeManaged,                   /* change_managed        */
    XtInheritInsertChild,                     /* insert_child          */
    XtInheritDeleteChild,                     /* delete_child          */
    NULL,                                     /* extension             */
  },

  {                                           /* constraint_class      */
    constraints,                              /* resources             */   
    XtNumber(constraints),                    /* num_resources         */   
    sizeof(XmpGridConstraintRec),             /* constraint_size       */
    ConstraintInitialize,                     /* constraint_initialize */
    NULL,                                     /* destroy               */   
    ConstraintSetValues,                      /* set_values            */   
    NULL,                                     /* extension             */
  },

  {                                           /* manager class         */
    XtInheritTranslations,                    /* translations          */
    syn_resources,                            /* syn_resources         */
    XtNumber(syn_resources),                  /* num_syn_resources     */
    syn_constraints,                          /* syn_constraint_resources */
    XtNumber(syn_constraints),                /* num_syn_constraint_resources */
    XmInheritParentProcess,                   /* parent_process        */
    NULL,                                     /* extension             */    
  },

  {                                           /* geometry class        */
    XmpInheritBitGravity,                     /* bit_gravity           */
    XmpInheritInitializePostHook,             /* initialize_post_hook  */
    XmpInheritSetValuesPostHook,              /* set_values_post_hook  */
    XmpInheritConstraintInitializePostHook,   /* constraint_initialize_post_hook*/
    XmpInheritConstraintSetValuesPostHook,    /* constraint_set_values_post_hook*/
    Size,                                     /* size                  */
    NULL,                                     /* extension             */    
  },

  {                                           /* grid class            */
    NULL,                                     /* extension             */    
  },        
};

externaldef(xmpgridwidgetclass) WidgetClass xmpGridWidgetClass =
			                        (WidgetClass) &xmpGridClassRec;


/************************************************************************
 *
 *  Initialize
 *
 ************************************************************************/
static void
Initialize (Widget request_w, Widget new_w, ArgList args, Cardinal* num_args)
{
    XmpGridWidget nw = (XmpGridWidget)new_w;

    if (Rows(nw) <= 0) {
	XmpWarning((Widget)nw, RowsZeroMessage);
	Rows(nw) = DefaultRows;
    }

    if (Columns(nw) <= 0) {
	XmpWarning((Widget)nw, ColumnsZeroMessage);
	Columns(nw) = DefaultColumns;
    }

    XmpGeometryInitialize(xmpGridWidgetClass, request_w, new_w, args, num_args);
}


/************************************************************************
 *
 *  Set Values
 *
 ************************************************************************/
static Boolean
SetValues (Widget old_w, Widget request_w, Widget new_w, ArgList args,
							Cardinal* num_args)
{
    XmpGridWidget cw = (XmpGridWidget)old_w;
    XmpGridWidget nw = (XmpGridWidget)new_w;
    Boolean redisplay = False;

    if (Rows(nw) <= 0) {
	XmpWarning((Widget)nw, RowsZeroMessage);
	Rows(nw) = Rows(cw);
    }

    if (Columns(nw) <= 0) {
	XmpWarning((Widget)nw, ColumnsZeroMessage);
	Columns(nw) = Columns(cw);
    }


    if (MarginWidth(nw) != MarginWidth(cw) ||
	MarginHeight(nw) != MarginHeight(cw) ||
	Rows(nw) != Rows(cw) ||
	Columns(nw) != Columns(cw)) {

	Reconfigure(nw) = True;
    }

    redisplay |= XmpGeometrySetValues(xmpGridWidgetClass,
					old_w, request_w, new_w, args, num_args);

    return (redisplay);
}


/************************************************************************
 *
 *  ConstraintInitialize
 *
 ************************************************************************/
static void
ConstraintInitialize (Widget rw, Widget nw, ArgList args, Cardinal* num_args)
{
    if (!XtIsRectObj (nw)) return;

    XmpGeometryConstraintInitialize(xmpGridWidgetClass, rw, nw, args, num_args);
}


/************************************************************************
 *
 *  ConstraintSetValues
 *
 ************************************************************************/
static Boolean
ConstraintSetValues (Widget cw, Widget rw, Widget nw, ArgList args,
							Cardinal* num_args)
{
    XmpGridWidget gw = (XmpGridWidget)XtParent(nw);
    Boolean redisplay = False;

    if (!XtIsRectObj (nw)) return (False);

    if (XtIsManaged(nw) && (
	GridMarginWidth(nw) != GridMarginWidth(cw) ||
	GridMarginHeight(nw) != GridMarginHeight(cw))) {

        ConstraintReconfigure(gw) = True;
    }

    redisplay |= XmpGeometryConstraintSetValues(xmpGridWidgetClass,
                                cw, rw, nw, args, num_args);
 
    return(redisplay);
}



/************************************************************************
 *
 *  Size
 *
 ************************************************************************/
static void
Size (Widget w, Dimension* gridWidth, Dimension* gridHeight)
{
    XmpGridWidget gw = (XmpGridWidget)w;
    Dimension maxWidth = 1;
    Dimension maxHeight = 1;
    int i;

    /* Calculate maximum cell size, including constraint margins */

    for (i=0; i < NumChildren(gw); i++) {
	Widget ic = Children(gw)[i];
	Dimension cb = BorderWidth(ic);
	Dimension cw, ch;

	if (!XtIsManaged(ic)) continue;

	XmpPreferredGeometry(ic, &cw, &ch);

	maxWidth = Max(maxWidth, cw+2*(cb+GridMarginWidth(ic)));
	maxHeight = Max(maxHeight, ch+2*(cb+GridMarginHeight(ic)));
    }

    /* Calculate overall size, including widget margins */

    *gridWidth = maxWidth * Columns(gw) + 2 * MarginWidth(gw);
    *gridHeight = maxHeight * Rows(gw) + 2 * MarginHeight(gw);
}



/************************************************************************
 *
 *  Resize 
 *
 ************************************************************************/
static void
Resize (Widget w)
{
    XmpGridWidget gw = (XmpGridWidget)w;
    Dimension mw = MarginWidth(gw);
    Dimension mh = MarginHeight(gw);
    Dimension cellWidth;
    Dimension cellHeight;
    int row, column;
    int i;

    cellWidth = XmpSubtract(Width(gw), 2u * mw, 0) / Columns(gw);
    cellHeight = XmpSubtract(Height(gw), 2u * mh, 0) / Rows(gw);

    for (i = 0, row = 0, column = 0; i < NumChildren(gw); i++) {
	Widget ic = Children(gw)[i];
	Dimension gmw = GridMarginWidth(ic);
	Dimension gmh = GridMarginHeight(ic);
	Position cx, cy;
	Dimension cw, ch;
	Dimension cb = BorderWidth(ic);

	if (!XtIsManaged(ic)) continue;

	cx = mw + column * cellWidth + gmw;
	cy = mh + row * cellHeight + gmh;
	cw = XmpSubtract(cellWidth, 2u * (gmw + cb), 1);
	ch = XmpSubtract(cellHeight, 2u * (gmh + cb), 1);

	XmpSetGeometry(ic,cx,cy,cw,ch,cb);

	if (++column == Columns(gw)) { column = 0; row++; }
    }
}



/************************************************************************
 *
 *  XmpCreateGrid
 *
 ************************************************************************/
Widget
XmpCreateGrid(Widget parent, char* name, Arg* args, Cardinal num_args)
{
    return(XtCreateWidget(name, xmpGridWidgetClass, parent, args, num_args));
}

